@@ -1,5 +1,6 @@
-from typing import List, Iterator, Optional, Literal, TypeVar, Generic, Type
+from typing import List, Iterator, Optional, Literal, TypeVar, Generic, Type, Union
 from enum import Enum
+from uuid import UUID
 from dataclasses import dataclass
 import pytest
 import sys
@@ -7,10 +8,11 @@ from contextlib import contextmanager
 import tempfile
 import shutil
 from pathlib import Path
-from radicli import Radicli, StaticRadicli, Arg
+from radicli import Radicli, StaticRadicli, Arg, get_arg, ArgparseArg
 from radicli.util import CommandNotFoundError, CliParserError
 from radicli.util import ExistingPath, ExistingFilePath, ExistingDirPath
-from radicli.util import ExistingFilePathOrDash
+from radicli.util import ExistingFilePathOrDash, DEFAULT_CONVERTERS
+from radicli.util import stringify_type
 
 
 @contextmanager
@@ -301,12 +303,14 @@ def test_cli_global_converters():
     assert ran
 
 
+_KindT = TypeVar("_KindT", bound=Union[str, int, float, Path])
+
+
+class CustomGeneric(Generic[_KindT]):
+    ...
+
+
 def test_cli_converters_generics():
-    _KindT = TypeVar("_KindT", bound=str)
-
-    class CustomGeneric(Generic[_KindT]):
-        ...
-
     converters = {CustomGeneric: lambda value: f"generic: {value}"}
     cli = Radicli(converters=converters)
     ran = False
@@ -827,3 +831,104 @@ def test_cli_static_roundtrip(capsys):
     static.run(["", "hello", "--help"])
     captured = capsys.readouterr().out
     assert not captured
+
+
+@pytest.mark.parametrize(
+    "arg_type",
+    [
+        str,
+        int,
+        float,
+        bool,
+        List[str],
+        Path,
+        ExistingPath,
+        ExistingFilePath,
+        ExistingDirPath,
+        ExistingFilePathOrDash,
+        Literal["a", "b", "c"],
+    ],
+)
+def test_static_deserialize_types(arg_type):
+    """Test that supported and built-in types are correctly deserialized from static"""
+    get_converter = lambda v: DEFAULT_CONVERTERS.get(v)
+    arg = get_arg(
+        "test", Arg("--test"), arg_type, orig_type=arg_type, get_converter=get_converter
+    )
+    arg_json = arg.to_static_json()
+    new_arg = ArgparseArg.from_static_json(arg_json)
+    assert new_arg.type == arg.type
+    assert new_arg.orig_type == stringify_type(arg.orig_type)
+    assert new_arg.has_converter == arg.has_converter
+    assert new_arg.action == arg.action
+
+
+@pytest.mark.parametrize(
+    "arg_type",
+    [
+        UUID,
+        List[str],
+        Optional[UUID],
+        Union[UUID, str],
+        CustomGeneric,
+        CustomGeneric[int],
+        CustomGeneric[str],
+        Optional[CustomGeneric],
+    ],
+)
+def test_static_deserialize_types_custom_deserialize(arg_type):
+    """Test deserialization with custom type deserializer"""
+
+    def split_string(text: str) -> List[str]:
+        return [t.strip() for t in text.split(",")] if text else []
+
+    def convert_uuid(value: str) -> UUID:
+        return UUID(value)
+
+    def convert_generic(value: str) -> str:
+        return f"generic: {value}"
+
+    converters = {
+        UUID: convert_uuid,
+        List[str]: split_string,
+        CustomGeneric: convert_generic,
+        CustomGeneric[str]: str,
+    }
+    get_converter = lambda v: converters.get(v)
+    # With converters set
+    arg = get_arg(
+        "test", Arg("--test"), arg_type, orig_type=arg_type, get_converter=get_converter
+    )
+    arg_json = arg.to_static_json()
+    new_arg = ArgparseArg.from_static_json(arg_json, converters=converters)
+    assert new_arg.type == arg.type
+    assert new_arg.orig_type == stringify_type(arg.orig_type)
+    assert new_arg.has_converter == arg.has_converter
+    assert new_arg.action == arg.action
+    # With no converters set
+    arg = get_arg(
+        "test", Arg("--test"), arg_type, orig_type=arg_type, get_converter=get_converter
+    )
+    arg_json = arg.to_static_json()
+    new_arg = ArgparseArg.from_static_json(arg_json)
+    assert new_arg.type is str
+    assert new_arg.orig_type == stringify_type(arg.orig_type)
+
+
+@pytest.mark.parametrize(
+    "arg_type,expected",
+    [
+        (str, "str"),
+        (bool, "bool"),
+        (Path, "Path"),
+        (List[int], "List[int]"),
+        (CustomGeneric, "CustomGeneric"),
+        (CustomGeneric[str], "CustomGeneric[str]"),
+        (UUID, "UUID"),
+        (shutil.rmtree, "rmtree"),
+        ("foo.bar", "foo.bar"),
+        (None, None),
+    ],
+)
+def test_stringify_type(arg_type, expected):
+    assert stringify_type(arg_type) == expected
