@@ -1,5 +1,6 @@
 from typing import Any, Callable, Iterable, Type, Union, Optional, Dict, Tuple
-from typing import List, Literal, NewType, get_args, get_origin, TypeVar, TypedDict
+from typing import List, Literal, NewType, get_args, get_origin, TypeVar
+from typing import TypedDict, cast
 from enum import Enum
 from dataclasses import dataclass
 from pathlib import Path
@@ -13,9 +14,11 @@ except ImportError:
     from collections import Iterable as IterableType  # type: ignore
 
 DEFAULT_PLACEHOLDER = argparse.SUPPRESS
-BASE_TYPES = [str, int, float, Path]
+BASE_TYPES_MAP = {"str": str, "int": int, "float": float, "Path": Path}
+BASE_TYPES = list(BASE_TYPES_MAP.values())
 ConverterType = Callable[[str], Any]
 ConvertersType = Dict[Union[Type, object], ConverterType]
+ArgTypeType = Optional[Union[Type, ConverterType]]
 _Exc = TypeVar("_Exc", bound=Exception, covariant=True)
 ErrorHandlerType = Callable[[Exception], Optional[int]]
 ErrorHandlersType = Dict[Type[_Exc], Callable[[_Exc], Optional[int]]]
@@ -31,6 +34,8 @@ class StaticArg(TypedDict):
     action: Optional[str]
     choices: Optional[List[str]]
     has_converter: bool
+    type: Optional[str]
+    orig_type: Optional[str]
 
 
 class StaticCommand(TypedDict):
@@ -51,6 +56,26 @@ class StaticData(TypedDict):
     subcommands: Dict[str, Dict[str, StaticCommand]]
 
 
+class SimpleFrozenDict(dict):
+    """Simplified implementation of a frozen dict, mainly used as default
+    function or method argument (for arguments that should default to empty
+    dictionary).
+    """
+
+    def __init__(self, *args, **kwargs) -> None:
+        super().__init__(*args, **kwargs)
+        self.error = "Can't write to frozen dict. This is likely an internal error."
+
+    def __setitem__(self, key, value):
+        raise NotImplementedError(self.error)
+
+    def pop(self, key: Any, default=None):
+        raise NotImplementedError(self.error)
+
+    def update(self, other, **kwargs):
+        raise NotImplementedError(self.error)
+
+
 class CliParserError(SystemExit):
     def __init__(self, message: str) -> None:
         self.message = message
@@ -107,8 +132,8 @@ class ArgparseArg:
 
     id: str
     arg: Arg
-    type: Optional[Union[Type, Callable[[str], Any]]] = None
-    orig_type: Optional[Union[Type, Callable[[str], Any]]] = None
+    type: ArgTypeType = None
+    orig_type: Union[ArgTypeType, str] = None
     default: Any = DEFAULT_PLACEHOLDER
     # We modify the help to add types so we store it twice to store old and new
     help: Optional[str] = None
@@ -117,7 +142,7 @@ class ArgparseArg:
     has_converter: bool = False
 
     @property
-    def display_type(self) -> Optional[Union[Type, Callable[[str], Any]]]:
+    def display_type(self) -> Union[ArgTypeType, str]:
         default_type = self.type if self.type is not None else self.orig_type
         return self.orig_type if self.has_converter else default_type
 
@@ -160,16 +185,22 @@ class ArgparseArg:
             if self.choices
             else None,
             "has_converter": self.has_converter,
+            "type": stringify_type(self.type),
+            "orig_type": stringify_type(self.orig_type),
         }
 
     @classmethod
-    def from_static_json(cls, data: StaticArg) -> "ArgparseArg":
+    def from_static_json(
+        cls,
+        data: StaticArg,
+        converters: ConvertersType = SimpleFrozenDict(),
+    ) -> "ArgparseArg":
         """Initialize the static argument from a JSON-serializable dict."""
         return ArgparseArg(
             id=data["id"],
             arg=Arg(data["option"], data["short"], help=data["orig_help"]),
-            type=str if not data["action"] else None,  # dummy, not used
-            orig_type=str if not data["action"] else None,  # dummy, not used
+            type=deserialize_type(data, converters),
+            orig_type=data["orig_type"],
             default=DEFAULT_PLACEHOLDER
             if data["default"] == DEFAULT_PLACEHOLDER
             else data["default"],
@@ -180,12 +211,42 @@ class ArgparseArg:
         )
 
 
+def deserialize_type(
+    data: StaticArg,
+    converters: ConvertersType = SimpleFrozenDict(),
+) -> ArgTypeType:
+    # Setting some common defaults here to handle out-of-the-box
+    if data["type"] is None:
+        return None
+    types_map = {**BASE_TYPES_MAP}
+    for value in DEFAULT_CONVERTERS.values():
+        types_map[stringify_type(value)] = value  # type: ignore
+    if data["action"] in ("store_true", "count"):  # special args with no type
+        return None
+    if data["type"] in types_map:
+        return types_map[data["type"]]
+    # Handle unknown types: we use the orig_type here, since this corresponds to
+    # what was actually set as an argument type hint
+    orig_type = data["orig_type"]
+    if orig_type is None:
+        return None
+    converters_map = {stringify_type(k): v for k, v in converters.items()}
+    if orig_type in converters_map:
+        return converters_map[orig_type]
+    # Hacky check for generics
+    if "[" in orig_type:
+        origin = orig_type.split("[", 1)[0]
+        if origin in converters_map:
+            return converters_map[orig_type.split("[", 1)[0]]
+    return str
+
+
 def get_arg(
     param: str,
     orig_arg: Arg,
     param_type: Any,
     *,
-    orig_type: Optional[Union[Type, Callable[[str], Any]]] = None,
+    orig_type: Union[ArgTypeType, str] = None,
     default: Optional[Any] = DEFAULT_PLACEHOLDER,
     get_converter: Optional[Callable[[Type], Optional[ConverterType]]] = None,
     skip_resolve: bool = False,
@@ -228,6 +289,7 @@ def get_arg(
                 param,
                 orig_arg,
                 arg_types[0],
+                orig_type=arg_types[0],
                 default=default,
                 get_converter=get_converter,
             )
@@ -280,15 +342,30 @@ def find_base_type(
     return default_type
 
 
-def format_type(arg_type: Any) -> str:
+def stringify_type(arg_type: Any) -> Optional[str]:
     """Get a pretty-printed string for a type."""
-    if isinstance(arg_type, type(NewType)) and hasattr(arg_type, "__supertype__"):  # type: ignore
-        return f"{arg_type.__name__} ({format_type(arg_type.__supertype__)})"  # type: ignore
+    if isinstance(arg_type, str) or arg_type is None:
+        return arg_type
     if hasattr(arg_type, "__name__"):
-        return arg_type.__name__
+        type_str = arg_type.__name__
+        args = get_args(arg_type)
+        if args:
+            # Built-in generic types are callables in Python 3.10+ so we want to
+            # preserve args here and stringify them, too
+            type_args = cast(List[str], [stringify_type(arg) for arg in args])
+            type_str = f"{type_str}[{', '.join(type_args)}]"
+        return type_str
     type_str = str(arg_type)
-    # Strip out typing for built-in types, leave path for custom
-    return type_str.replace("typing.", "")
+    split_type = type_str.rsplit(".", 1)
+    return split_type[1] if len(split_type) == 2 else type_str
+
+
+def format_type(arg_type: Any) -> Optional[str]:
+    """Get a pretty-printed string for a type."""
+    type_str = stringify_type(arg_type)
+    if isinstance(arg_type, type(NewType)) and hasattr(arg_type, "__supertype__"):  # type: ignore
+        return f"{type_str} {arg_type.__name__}"  # type: ignore
+    return type_str
 
 
 def join_strings(*strings: Optional[str], char: str = " ") -> str:
@@ -393,23 +470,3 @@ DEFAULT_CONVERTERS: ConvertersType = {
     ExistingDirPathOrDash: convert_existing_dir_path_or_dash,
     PathOrDash: convert_path_or_dash,
 }
-
-
-class SimpleFrozenDict(dict):
-    """Simplified implementation of a frozen dict, mainly used as default
-    function or method argument (for arguments that should default to empty
-    dictionary).
-    """
-
-    def __init__(self, *args, **kwargs) -> None:
-        super().__init__(*args, **kwargs)
-        self.error = "Can't write to frozen dict. This is likely an internal error."
-
-    def __setitem__(self, key, value):
-        raise NotImplementedError(self.error)
-
-    def pop(self, key: Any, default=None):
-        raise NotImplementedError(self.error)
-
-    def update(self, other, **kwargs):
-        raise NotImplementedError(self.error)
